///////////////////////////////////////////////////////////////////////////
//                                                                       //
//  Class: Enriched_polyhedron                                           //
//                                                                       //
///////////////////////////////////////////////////////////////////////////

#ifndef _POLYGON_MESH_
#define _POLYGON_MESH_

// CGAL stuff
#include <CGAL/Cartesian.h>
#include <CGAL/Polyhedron_3.h>
#include <list>

// a refined facet with a normal and a tag
template <class Refs, class T, class P, class Norm>
class Enriched_facet : public CGAL::HalfedgeDS_face_base<Refs, T>
{
  // tag
  int m_tag; 
  
  // normal
  Norm m_normal;

public:

  // life cycle
  // no constructors to repeat, since only
  // default constructor mandatory

	const Norm & getUnitNormal() const { return m_normal; }
  void setNormal(Norm n) { m_normal = n; }

  Enriched_facet()
  {
  }

  // tag
  const int&  tag() { return m_tag; }
  void tag(const int& t)  { m_tag = t; }

  
  // normal
  typedef Norm Normal_3;
  Normal_3& normal() { return m_normal; }
  const Normal_3& normal() const { return m_normal; }

  	
};

// a refined halfedge with a general tag and 
// a binary tag to indicate wether it belongs 
// to the control mesh or not
template <class Refs, class Tprev, class Tvertex, class Tface, class Norm>
class Enriched_halfedge : public CGAL::HalfedgeDS_halfedge_base<Refs,Tprev,Tvertex,Tface>
{
private:

  // tag
  int m_tag; 

  

public:

 

  // tag
  const int& tag() const { return m_tag;  }
  int& tag() { return m_tag;  }
  void tag(const int& t)  { m_tag = t; }

  // control edge 
  
};



// a refined vertex with a normal and a tag
template <class Refs, class T, class P, class Vec>
class Enriched_vertex : public CGAL::HalfedgeDS_vertex_base<Refs, T, P>
{
  // tag
  int m_tag; 
 
 

  // Variables pour la courbure
	
	//scalaire MinCurvature
	double m_Kmin;
	//scalaire MaxCurvature
	double m_Kmax;

	double m_Kg;
	double m_Kh;

	//vecteur courbures principales
	Vec m_VKmin;
	Vec m_VKmax;


  // normal
  Vec m_normal;

public:

	float CourbureCoVariance;
	
  // life cycle
  Enriched_vertex()  {}
  // repeat mandatory constructors
  Enriched_vertex(const P& pt)
    : CGAL::HalfedgeDS_vertex_base<Refs, T, P>(pt)
  {
  }

  // normal
  typedef Vec Normal_3;
  Normal_3& normal() { return m_normal; }
  const Normal_3& normal() const { return m_normal; }

  // tag
 
   int& tag() {  return m_tag; }
  const int& tag() const {  return m_tag; }
  void tag(const int& t)  { m_tag = t; }

  // mean curvature
	double& Kh() {	return m_Kh;	}
	const double& Kh() const {	return m_Kh;	}
	void Kh(const double& Kh)	{	m_Kh	=	Kh; }

  double& Kg() {	return m_Kg;	}
	const double& Kg() const {	return m_Kg;	}
	void Kg(const double& kg)	{	m_Kg	=	kg; }
	double KgNrm(){return fabs(m_Kg);}

	// min curvature
	double& Kmin() {	return m_Kmin;	}
	const double& Kmin() const {	return m_Kmin;	}
	void Kmin(const double& Kmin)	{	m_Kmin	=	Kmin; }
	double KminNrm(){return fabs(m_Kmin);}

	// max curvature
	double& Kmax() {	return m_Kmax;	}
	const double& Kmax() const {	return m_Kmax;	}
	void Kmax(const double& Kmax)	{	m_Kmax	=	Kmax; }

	
	Vec& VKmin() {	return m_VKmin;	}
	const Vec& VKmin() const {	return m_VKmin;	}
	void VKmin(const Vec& VKmin)	{	m_VKmin	=	VKmin; }


	Vec& VKmax() {	return m_VKmax;	}
	const Vec& VKmax() const {	return m_VKmax;	}
	void VKmax(const Vec& VKmax)	{	m_VKmax	=	VKmax; }

	
	
	double CourbureMoyenne;
	double CourbureVariance;
	


  
};

// A redefined items class for the Polyhedron_3 
// with a refined vertex class that contains a 
// member for the normal vector and a refined
// facet with a normal vector instead of the 
// plane equation (this is an alternative 
// solution instead of using 
// Polyhedron_traits_with_normals_3).

struct Enriched_items : public CGAL::Polyhedron_items_3
{
    // wrap vertex
    template <class Refs, class Traits>
    struct Vertex_wrapper
    {
        typedef typename Traits::Point_3  Point;
        typedef typename Traits::Vector_3 Normal;
        typedef Enriched_vertex<Refs,
                          CGAL::Tag_true,
                          Point,
                          Normal> Vertex;
    };

    // wrap face
    template <class Refs, class Traits>
    struct Face_wrapper
    {
        typedef typename Traits::Point_3  Point;
        typedef typename Traits::Vector_3 Normal;
        typedef Enriched_facet<Refs,
                         CGAL::Tag_true,
                         Point,
                         Normal> Face;
    };

    // wrap halfedge
    template <class Refs, class Traits>
    struct Halfedge_wrapper
    {
        typedef typename Traits::Vector_3 Normal;
        typedef Enriched_halfedge<Refs,
                            CGAL::Tag_true,
                            CGAL::Tag_true,
                            CGAL::Tag_true,
                            Normal> Halfedge;
    };
};

//*********************************************************
template <class kernel, class items>
class Enriched_polyhedron : public CGAL::Polyhedron_3<kernel,items>
{
public :
	
  typedef typename kernel::FT FT;
  typedef typename kernel::Point_3 Point;
  typedef typename kernel::Vector_3 Vector;
  typedef typename kernel::Iso_cuboid_3 Iso_cuboid;
   typedef typename CGAL::Polyhedron_3<kernel,items> poly3;


private :

  // bounding box
  Iso_cuboid m_bbox;

  //courbure extrema
  double m_MinNrmMinCurvature;
double m_MaxNrmMinCurvature;
double m_MinNrmMaxCurvature;
double m_MaxNrmMaxCurvature;
double m_MinNrmRoughCurvature;
double m_MaxNrmRoughCurvature;
double m_MinNrmMeanCurvature;
	double m_MaxNrmMeanCurvature;
	double m_MinNrmGaussCurvature;
	double m_MaxNrmGaussCurvature;
	double m_MinNrmSalCurvature;
	double m_MaxNrmSalCurvature;






public :


	char m_Name[256];	

//pour la fusion de rgion
double AireTotale;


  double& MinNrmMeanCurvature() {	return m_MinNrmMeanCurvature;	}
	const double& MinNrmMeanCurvature() const {	return m_MinNrmMeanCurvature;	}
	void MinNrmMeanCurvature(const double& t)	{	m_MinNrmMeanCurvature	=	t; }

	double& MaxNrmMeanCurvature() {	return m_MaxNrmMeanCurvature;	}
	const double& MaxNrmMeanCurvature() const {	return m_MaxNrmMeanCurvature;	}
	void MaxNrmMeanCurvature(const double& t)	{	m_MaxNrmMeanCurvature	=	t; }


  double& MinNrmMinCurvature() {	return m_MinNrmMinCurvature;	}
	const double& MinNrmMinCurvature() const {	return m_MinNrmMinCurvature;	}
	void MinNrmMinCurvature(const double& t)	{	m_MinNrmMinCurvature	=	t; }

	double& MaxNrmMinCurvature() {	return m_MaxNrmMinCurvature;	}
	const double& MaxNrmMinCurvature() const {	return m_MaxNrmMinCurvature;	}
	void MaxNrmMinCurvature(const double& t)	{	m_MaxNrmMinCurvature	=	t; }

	double& MinNrmMaxCurvature() {	return m_MinNrmMaxCurvature;	}
	const double& MinNrmMaxCurvature() const {	return m_MinNrmMaxCurvature;	}
	void MinNrmMaxCurvature(const double& t)	{	m_MinNrmMaxCurvature	=	t; }

	double& MaxNrmMaxCurvature() {	return m_MaxNrmMaxCurvature;	}
	const double& MaxNrmMaxCurvature() const {	return m_MaxNrmMaxCurvature;	}
	void MaxNrmMaxCurvature(const double& t)	{	m_MaxNrmMaxCurvature	=	t; }

	double& MaxNrmRoughCurvature() {	return m_MaxNrmRoughCurvature;	}
	const double& MaxNrmRoughCurvature() const {	return m_MaxNrmRoughCurvature;	}
	void MaxNrmRoughCurvature(const double& t)	{	m_MaxNrmRoughCurvature	=	t; }

	double& MinNrmRoughCurvature() {	return m_MinNrmRoughCurvature;	}
	const double& MinNrmRoughCurvature() const {	return m_MinNrmRoughCurvature;	}
	void MinNrmRoughCurvature(const double& t)	{	m_MinNrmRoughCurvature	=	t; }

	double& MaxNrmSalCurvature() {	return m_MaxNrmSalCurvature;	}
	const double& MaxNrmSalCurvature() const {	return m_MaxNrmSalCurvature;	}
	void MaxNrmSalCurvature(const double& t)	{	m_MaxNrmSalCurvature	=	t; }

	double& MinNrmSalCurvature() {	return m_MinNrmSalCurvature;	}
	const double& MinNrmSalCurvature() const {	return m_MinNrmSalCurvature;	}
	void MinNrmSalCurvature(const double& t)	{	m_MinNrmSalCurvature	=	t; }

	double& MinNrmGaussCurvature() {	return m_MinNrmGaussCurvature;	}
	const double& MinNrmGaussCurvature() const {	return m_MinNrmGaussCurvature;	}
	void MinNrmGaussCurvature(const double& t)	{	m_MinNrmGaussCurvature	=	t; }

	double& MaxNrmGaussCurvature() {	return m_MaxNrmGaussCurvature;	}
	const double& MaxNrmGaussCurvature() const {	return m_MaxNrmGaussCurvature;	}
	void MaxNrmGaussCurvature(const double& t)	{	m_MaxNrmGaussCurvature	=	t; }

	

	char* Name() {	return m_Name;	}
	const char* Name() const {	return m_Name;	}
	void Name(char* _Name)	{	strcpy( m_Name,_Name);}


  // normals (per facet, then per vertex)
  void compute_normals_per_facet()
  {
    std::for_each(facets_begin(),facets_end(),Facet_normal());
  }
  void compute_normals_per_vertex()
  {
    std::for_each(vertices_begin(),vertices_end(),Vertex_normal());
  }
  void compute_normals()
  {
    compute_normals_per_facet();
    compute_normals_per_vertex();
  }

  // bounding box
  Iso_cuboid& bbox() { return m_bbox; }
  const Iso_cuboid bbox() const { return m_bbox; }

  // compute bounding box
  void compute_bounding_box()
  {
    if(size_of_vertices() == 0)
    {
     // ASSERT(false);
      return;
    }

    FT xmin,xmax,ymin,ymax,zmin,zmax;
    Vertex_iterator pVertex = vertices_begin();
    xmin = xmax = pVertex->point().x();
    ymin = ymax = pVertex->point().y();
    zmin = zmax = pVertex->point().z();
    for(;
        pVertex !=  vertices_end();
        pVertex++)
    {
      const Point& p = pVertex->point();

      xmin =  std::min(xmin,p.x());
      ymin =  std::min(ymin,p.y());
      zmin =  std::min(zmin,p.z());

      xmax =  std::max(xmax,p.x());
      ymax =  std::max(ymax,p.y());
      zmax =  std::max(zmax,p.z());
    }
    m_bbox = Iso_cuboid(xmin,ymin,zmin,
                        xmax,ymax,zmax);
  }

  // bounding box
  FT xmin() { return m_bbox.xmin(); }
  FT xmax() { return m_bbox.xmax(); }
  FT ymin() { return m_bbox.ymin(); }
  FT ymax() { return m_bbox.ymax(); }
  FT zmin() { return m_bbox.zmin(); }
  FT zmax() { return m_bbox.zmax(); }

  // copy bounding box
  void copy_bounding_box(Enriched_polyhedron<kernel,items> *pMesh)
  {
    m_bbox = pMesh->bbox();
  }

  // degree of a face
  static unsigned int degree(Facet_handle pFace)
  {
    return CGAL::circulator_size(pFace->facet_begin());    
  }

  

  // valence of a vertex
  static unsigned int valence(Vertex_handle pVertex)
  {
    return CGAL::circulator_size(pVertex->vertex_begin());
  }

  // check wether a vertex is on a boundary or not
  static bool is_border(Vertex_handle pVertex)
  {
    Halfedge_around_vertex_circulator pHalfEdge = pVertex->vertex_begin();
    if(pHalfEdge == NULL) // isolated vertex
      return true;
    Halfedge_around_vertex_circulator d = pHalfEdge;
    CGAL_For_all(pHalfEdge,d)
      if(pHalfEdge->is_border())
        return true;
    return false;
  }

  // get any border halfedge attached to a vertex
  Halfedge_handle get_border_halfedge(Vertex_handle pVertex)
  {
    Halfedge_around_vertex_circulator pHalfEdge = pVertex->vertex_begin();
    Halfedge_around_vertex_circulator d = pHalfEdge;
    CGAL_For_all(pHalfEdge,d)
      if(pHalfEdge->is_border())
        return pHalfEdge;
    return NULL;
  }

  // get any other border	halfedge attached	to a vertex
	Halfedge_handle	get_other_border_halfedge(Vertex_handle	pVertex,Halfedge_handle HE)
	{
		Halfedge_around_vertex_circulator	pHalfEdge	=	pVertex->vertex_begin();
		Halfedge_around_vertex_circulator	d	=	pHalfEdge;
		CGAL_For_all(pHalfEdge,d)
			if(pHalfEdge->is_border())
				if(pHalfEdge->opposite()->vertex()!=HE->opposite()->vertex())
				return pHalfEdge;
		return NULL;
	}
  // tag all halfedges
  void tag_halfedges(const int tag)
  {
    for(Halfedge_iterator pHalfedge = halfedges_begin();
        pHalfedge != halfedges_end();
        pHalfedge++)
      pHalfedge->tag(tag);
  }

  // tag all facets
  void tag_facets(const int tag)
  {
    for(Facet_iterator pFacet = facets_begin();
        pFacet  != facets_end();
        pFacet++)
      pFacet->tag(tag);
  }

  // set index for all vertices
  void set_index_vertices()
  {
    int index = 0;
    for(Vertex_iterator pVertex = vertices_begin();
        pVertex != vertices_end();
        pVertex++)
      pVertex->tag(index++);
  }


  // compute facet center
  void compute_facet_center(Facet_handle pFace,
                            Point& center)
  {
    Halfedge_around_facet_circulator pHalfEdge = pFace->facet_begin();
    Halfedge_around_facet_circulator end = pHalfEdge;
    Vector vec(0.0,0.0,0.0);
    int degree = 0;
    CGAL_For_all(pHalfEdge,end)
    {
      vec = vec + (pHalfEdge->vertex()->point()-CGAL::ORIGIN);
      degree++;
    }
    center = CGAL::ORIGIN + (vec/(kernel::FT)degree);
  }

  // compute average edge length around a vertex
  FT average_edge_length_around(Vertex_handle pVertex)
  {
    FT sum = 0.0;
    Halfedge_around_vertex_circulator pHalfEdge = pVertex->vertex_begin();
    Halfedge_around_vertex_circulator end = pHalfEdge;
    Vector vec(0.0,0.0,0.0);
    int degree = 0;
    CGAL_For_all(pHalfEdge,end)
    {
      Vector vec = pHalfEdge->vertex()->point()-
                   pHalfEdge->opposite()->vertex()->point();
      sum += std::sqrt(vec*vec);
      degree++;
    }
    return sum / (FT) degree;
  }

  double getMaxDim()
  {
	compute_bounding_box();
		
		double max=xmax()-xmin();
		if(ymax()-ymin()>max)
			max = ymax()-ymin();
		if(zmax()-zmin()>max)
			max = zmax()-zmin();

		return max;

  }
  void Normalise()
	{
		
		compute_bounding_box();
		
		double max=xmax()-xmin();
		if(ymax()-ymin()>max)
			max = ymax()-ymin();
		if(zmax()-zmin()>max)
			max = zmax()-zmin();

		//double diag=sqrt((xmax()-xmin())*(xmax()-xmin())+(ymax()-ymin())*(ymax()-ymin())+(zmax()-zmin())*(zmax()-zmin()));
		Vector center((xmax()+xmin())/2.,(ymax()+ymin())/2.,(zmax()+zmin())/2.);
		for(Vertex_iterator	pVertex	=	vertices_begin();pVertex!= vertices_end();pVertex++)
		{
			pVertex->point()=pVertex->point()-center;//Point(pVertex->point().x()-xmin(),pVertex->point().y()-ymin(),pVertex->point().z()-zmin());
			
		}

		for(Vertex_iterator	pVertex	=	vertices_begin();pVertex!= vertices_end();pVertex++)
		{
			pVertex->point()=Point(pVertex->point().x()/max,pVertex->point().y()/max,pVertex->point().z()/max);
			
		}

	}

  

  // write in obj file format (OBJ).
  void write_obj(char *pFilename,
                  int incr  = 1) // 1-based by default
  {
    std::ofstream stream(pFilename);

    // output vertices
    for(Point_iterator pPoint = points_begin();
        pPoint != points_end(); 
        pPoint++) 
      stream << 'v' << ' ' << pPoint->x() << ' ' <<
                              pPoint->y() << ' ' <<
                              pPoint->z() << std::endl;

    // precompute vertex indices
    this->set_index_vertices(); 

    // output facets
    for(Facet_iterator pFacet = facets_begin();
        pFacet != facets_end(); 
        pFacet++) 
    {
      stream << 'f';
      Halfedge_around_facet_circulator pHalfedge = pFacet->facet_begin();
      do 
        stream << ' ' << pHalfedge->vertex()->tag()+incr;
      while(++pHalfedge != pFacet->facet_begin());
      stream << std::endl;
    }  
  }

  


  unsigned int nb_boundaries()
  {
    unsigned int nb = 0;
    tag_halfedges(0);
    for(Halfedge_iterator he = halfedges_begin();
        he != halfedges_end();
        he++)
    {
      if(he->is_border() && he->tag() == 0)
      {
        nb++;
        Halfedge_handle curr = he;
        do
        {
          curr  = curr->next();
          curr->tag(1);
        }
        while(curr != he);
      }
    }
    return nb;
  }

  // tag component 
  void tag_component(Facet_handle pSeedFacet,
                     const int tag_free,
                     const int tag_done)
{
    pSeedFacet->tag(tag_done);
    std::list<Facet_handle> facets;
    facets.push_front(pSeedFacet);
    while(!facets.empty())
    {
      Facet_handle pFacet = facets.front();
      facets.pop_front();
      pFacet->tag(tag_done);
      Halfedge_around_facet_circulator pHalfedge = pFacet->facet_begin();
      Halfedge_around_facet_circulator end = pHalfedge;
      CGAL_For_all(pHalfedge,end)
      {
        Facet_handle pNFacet = pHalfedge->opposite()->facet();
        if(pNFacet != NULL && pNFacet->tag() == tag_free)
        {
          facets.push_front(pNFacet);
          pNFacet->tag(tag_done);
        }
      }
    }
}

  // count #components
  unsigned int nb_components()
  {
    unsigned int nb = 0;
    tag_facets(0);
    for(Facet_iterator pFacet = facets_begin();
        pFacet != facets_end();
        pFacet++)
    {
      if(pFacet->tag() == 0)
      {
        nb++;
        tag_component(pFacet,0,1);
      }
    }
    return nb;
  }

  // compute the genus
  // V - E + F + B = 2 (C - G)
  // C -> #connected components
  // G : genus
  // B : #boundaries
  int genus()
  {
    int c = nb_components();
    int b = nb_boundaries();
    int v = size_of_vertices();
    int e = size_of_halfedges()/2;
    int f = size_of_facets();
    return genus(c,v,f,e,b);
  }
  int genus(int c,
            int v,
            int f,
            int e,
            int b)
  {
    return (2*c+e-b-f-v)/2;
  }

  double AreaFacetTriangle(Facet_iterator &f)
	{
		Halfedge_around_facet_circulator pHalfedge = f->facet_begin();
		Point P = pHalfedge->vertex()->point();
		Point Q = pHalfedge->next()->vertex()->point();
		Point R = pHalfedge->next()->next()->vertex()->point();

		Vector PQ=Q-P;
		Vector PR=R-P;
		Vector QR=R-Q;

	
		Vector normal	=	CGAL::cross_product(PQ,QR);
		double area=0.5*sqrt(normal*normal);

		return area;

	}

	double HalfEdgeLenght(Halfedge_iterator &pHalfedge)
	{
		
		Point P = pHalfedge->vertex()->point();
		Point Q = pHalfedge->opposite()->vertex()->point();

		Vector PQ=Q-P;

		double lenght;
		lenght	=	sqrt(PQ*PQ);

		return lenght;

	}

	void tag_vertices(const	int	tag)
	{
		for(Vertex_iterator	pVertex	=	vertices_begin();
				pVertex	!= vertices_end();
				pVertex++)
			pVertex->tag(tag);
	}
	void compute_Courbure_NormalCycle(bool IsGeod, double radius)
	{
		Normal_cycle<My_polyhedron<My_kernel,My_items> > estimateur;
		estimateur.principal_curvature(*this,IsGeod,radius);
		
		
		
		Vertex_iterator	pVertex	=	vertices_begin();
		

		MinNrmMinCurvature ( pVertex->Kmin());
		MaxNrmMinCurvature ( pVertex->Kmin());

		MinNrmMaxCurvature ( pVertex->Kmax());
		MaxNrmMaxCurvature ( pVertex->Kmax());

		
		
		for(;pVertex !=	vertices_end();pVertex++)
		{
			
			

			MinNrmMinCurvature	(std::min(MinNrmMinCurvature(),pVertex->Kmin()));
			MaxNrmMinCurvature(std::max(MaxNrmMinCurvature(),pVertex->Kmin()));

			MinNrmMaxCurvature(std::min(MinNrmMaxCurvature(),pVertex->Kmax()));
			MaxNrmMaxCurvature(std::max(MaxNrmMaxCurvature(),pVertex->Kmax()));

			
			
		}
		m_IsCourbureCalculated=true;
		m_IsCourbureDirectionsCalculated=true;
		

	}

	void KmaxKmean(double coef)
	{
		double max=0;
		double min=10000000;
		for(Vertex_iterator	pVertex	=	vertices_begin();pVertex!= vertices_end();pVertex++)
		{
			double kmax=pVertex->Kmax()*coef;
			double kmin=pVertex->Kmin()*coef;

			pVertex->Kmax((kmax+kmin)/2.);

			if(pVertex->Kmax()>max)
				max=pVertex->Kmax();
			if(pVertex->Kmax()<min)
				min=pVertex->Kmax();

		}

		MinNrmMaxCurvature(min);
		MaxNrmMaxCurvature(max);

		//GaussianNormalisationKmax();

	}


	
  


};

// compute facet normal 
struct Facet_normal // (functor)
{
  template <class Facet>
  void operator()(Facet& f)
  {
    typename Facet::Normal_3 sum = CGAL::NULL_VECTOR;
    typename Facet::Halfedge_around_facet_circulator h = f.facet_begin();
    do
    {
      typename Facet::Normal_3 normal = CGAL::cross_product(
        h->next()->vertex()->point() - h->vertex()->point(),
        h->next()->next()->vertex()->point() - h->next()->vertex()->point());
      double sqnorm = normal * normal;
      if(sqnorm != 0)
        normal = normal / (float)std::sqrt(sqnorm);
      sum = sum + normal;
    }
    while(++h != f.facet_begin());
    float sqnorm = sum * sum;
    if(sqnorm != 0.0)
      f.normal() = sum / std::sqrt(sqnorm);
    else
    {
      f.normal() = CGAL::NULL_VECTOR;
     // TRACE("degenerate face\n");
    }
  }
};


// compute vertex normal 
struct Vertex_normal // (functor)
{
    template <class Vertex>
    void operator()(Vertex& v)
    {
        typename Vertex::Normal_3 normal = CGAL::NULL_VECTOR;
        Vertex::Halfedge_around_vertex_const_circulator pHalfedge = v.vertex_begin();
        Vertex::Halfedge_around_vertex_const_circulator begin = pHalfedge;
        CGAL_For_all(pHalfedge,begin) 
          if(!pHalfedge->is_border())
            normal = normal + pHalfedge->facet()->normal();
        float sqnorm = normal * normal;
        if(sqnorm != 0.0f)
          v.normal() = normal / (float)std::sqrt(sqnorm);
        else
          v.normal() = CGAL::NULL_VECTOR;
    }
};

#endif // _POLYGON_MESH_
